#include <cstdlib>
#include <cstring>
#include <cmath>
#include "OBJModel.h"

#include <fstream>
#include <sstream>

using namespace std;

OBJModel::OBJModel(const char *fileName, float scale) {

  _scale = scale;

  int sz = strlen(fileName);
  char *mtlFileName = (char *)malloc(sz+1); // termination character
  strcpy(mtlFileName,fileName);
  mtlFileName[sz-3] = 'm'; mtlFileName[sz-2] = 't'; mtlFileName[sz-1] = 'l';
  FILE* mtlFile = fopen(mtlFileName, "r");
  if (mtlFile!=NULL) {
    fclose(mtlFile);
    // read file
    readMTLFile(mtlFileName);
  } else
    printf("mtl file open failed!\n");

  free(mtlFileName);


  FILE* objFile = fopen(fileName, "r");
  if (objFile!=NULL) {
    // read file
    readFile(objFile);
    fclose(objFile);
  } else
    printf("obj file open failed!\n");

}


OBJModel::~OBJModel() {

  delete[] _positions;
  delete[] _verticesGL;

}


float OBJModel::maximalDistance(const float *T1, const float *R_mat1, const float *T2, const float *R_mat2) {

  float maxDist = -1.0f;
  float p1[3], p2[3];

  for(int vertex=0;vertex<_nPositions;vertex++) {

    for(int i=0;i<3;i++) {
      p1[i] = 0.0f;
      p2[i] = 0.0f;
    }

    // transform positions
    for(int r=0;r<3;r++) {
      for(int i=0;i<3;i++) {
        p1[r] += R_mat1[r*3+i] * _positions[3*vertex+i];
        p2[r] += R_mat2[r*3+i] * _positions[3*vertex+i];
      }
      p1[r] += T1[r];
      p2[r] += T2[r];
    }

    // compute distance squared
    for(int i=0;i<3;i++)
      p1[i] -= p2[i];

    float dist = p1[0]*p1[0] + p1[1]*p1[1] + p1[2]*p1[2];

    maxDist = (dist > maxDist) ? dist : maxDist;

  }

  if (maxDist == -1.0f) {
    return(1e6f);
  } else
    return(sqrt(maxDist));

}


void OBJModel::readMTLFile(const char *mtlFileName) {

  string imgFileName;

  string line;
  ifstream mtlFile(mtlFileName);
  if (mtlFile.is_open())
  {
    while ( mtlFile.good() )
    {
      getline (mtlFile,line);
      istringstream iss(line);
      while(iss) {
        string sub;
        iss >> sub;
        if (sub == "map_Kd")
          iss >> imgFileName;
      }
    }
    mtlFile.close();
  }

  else cout << "Unable to open file";

  // add path to image filename
  string tmp = mtlFileName;
  size_t found = tmp.find_last_of("/");
  _imgFileName = tmp.substr(0,found) + "/" + imgFileName;

//  cout << _imgFileName << endl;

}


void OBJModel::readFile(FILE *objFile) {

  int nVertices = 0, nTexCoords = 0, nNormals = 0, nTriangles = 0;

  rewind(objFile);

  // first pass to count vertices, texture coords, normals and faces (triangles)

  char buf[128];

  while(fscanf(objFile, "%s", buf) != EOF) {

    switch(buf[0]) {

      case 'f': nTriangles++; fgets(buf, sizeof(buf), objFile); break;

      case 'v': // vertex, texture coord, or normal
        switch (buf[1]) {
          case '\0': nVertices++; fgets(buf, sizeof(buf), objFile); break;
          case 't': nTexCoords++; fgets(buf, sizeof(buf), objFile); break;
          case 'n': nNormals++; fgets(buf, sizeof(buf), objFile); break;
        }
        break;

      default: fgets(buf, sizeof(buf), objFile); break;

      }
  }

//  printf("_nVertices %d, _nTexCoords %d, _nNormals %d, _nTriangles %d\n",_nVertices, _nTexCoords, _nNormals, _nTriangles);


  // allocate space for primitives
  float *vertices = (float *)malloc(3*nVertices*sizeof(float));
  float *texCoords = (float *)malloc(2*nTexCoords*sizeof(float));
  float *normals = (float *)malloc(3*nNormals*sizeof(float));
  int *triangleINDs = (int *)malloc(9*nTriangles*sizeof(int));;

  // second pass for reading

  rewind(objFile);

  int vertices_cnt = 0, texCoords_cnt = 0, normals_cnt = 0, triangleINDs_cnt = 0;
  int *tmpI;
  float *tmpF;

  while(fscanf(objFile, "%s", buf) != EOF) {

    switch(buf[0]) {

      case 'f':
        tmpI = &triangleINDs[9*triangleINDs_cnt];
        fscanf(objFile, "%d/%d/%d %d/%d/%d %d/%d/%d", &tmpI[0], &tmpI[1], &tmpI[2], &tmpI[3], &tmpI[4], &tmpI[5], &tmpI[6], &tmpI[7], &tmpI[8]);
        triangleINDs_cnt++;
        break;

      case 'v': // vertex, texture coord, or normal
        switch (buf[1]) {
          case '\0':
            tmpF = &vertices[3*vertices_cnt];
            fscanf(objFile, "%f %f %f", &tmpF[0], &tmpF[1], &tmpF[2]);
            vertices_cnt++;
            break;
          case 't':
            tmpF = &texCoords[2*texCoords_cnt];
            fscanf(objFile, "%f %f", &tmpF[0], &tmpF[1]);
            texCoords_cnt++;
            break;
          case 'n':
            tmpF = &normals[3*normals_cnt];
            fscanf(objFile, "%f %f %f", &tmpF[0], &tmpF[1], &tmpF[2]);
            normals_cnt++;
            break;
        }
        break;

      default: fgets(buf, sizeof(buf), objFile); break;

      }
  }

//  for(int i=0;i<_nVertices;i++) {
//    printf("v  ");
//    for(int j=0;j<3;j++)
//      printf("%2.6f ",vertices[3*i+j]);
//    printf("\n");
//  }
//
//  for(int i=0;i<_nTexCoords;i++) {
//    printf("t  ");
//    for(int j=0;j<2;j++)
//      printf("%2.6f ",texCoords[2*i+j]);
//    printf("\n");
//  }
//
//  for(int i=0;i<_nNormals;i++) {
//    printf("n  ");
//    for(int j=0;j<3;j++)
//      printf("%2.6f ",normals[3*i+j]);
//    printf("\n");
//  }
//
//  for(int i=0;i<_nTriangles;i++) {
//    printf("tr ");
//    for(int j=0;j<9;j++)
//      printf("%d/",triangleINDs[9*i+j]);
//    printf("\n");
//  }



  // transform to openGL compatible format (careful with obj's 1-based indices)
  _nVerticesGL = nTriangles*3;
  _verticesGL = new VertexOBJ[_nVerticesGL];

  for(int vertex=0;vertex<_nVerticesGL;vertex++) {

    int *IND = &triangleINDs[3*vertex];

    _verticesGL[vertex].Position[0] = _scale*vertices[3*(IND[0]-1)];
    _verticesGL[vertex].Position[1] = _scale*vertices[3*(IND[0]-1)+1];
    _verticesGL[vertex].Position[2] = _scale*vertices[3*(IND[0]-1)+2];

    _verticesGL[vertex].Texture[0] = texCoords[2*(IND[1]-1)];
    _verticesGL[vertex].Texture[1] = texCoords[2*(IND[1]-1)+1];

    _verticesGL[vertex].Normal[0] = normals[3*(IND[2]-1)];
    _verticesGL[vertex].Normal[1] = normals[3*(IND[2]-1)+1];
    _verticesGL[vertex].Normal[2] = normals[3*(IND[2]-1)+2];

  }


//  for(int vertex=0;vertex<_nVerticesGL;vertex++) {
//    printf("v %+2.2f %+2.2f %+2.2f -- t %+2.2f %+2.2f -- n %+2.2f %+2.2f %+2.2f\n",_verticesGL[vertex].Position[0],_verticesGL[vertex].Position[1],_verticesGL[vertex].Position[2],_verticesGL[vertex].Texture[0],_verticesGL[vertex].Texture[1],_verticesGL[vertex].Normal[0],_verticesGL[vertex].Normal[1],_verticesGL[vertex].Normal[2]);
//  }


  // transform positions to real-world coordinate system (scaled and z-axi away from observer)
  _nPositions = nVertices;
  _positions = (float *)malloc(3*_nPositions*sizeof(float));

  for(int vertex=0;vertex<_nPositions;vertex++) {
    _positions[vertex*3] = _scale*vertices[3*vertex];
    _positions[vertex*3+1] = _scale*vertices[3*vertex+1];
    _positions[vertex*3+2] = -_scale*vertices[3*vertex+2]; // flip z
  }


  // clear primitives
  free(triangleINDs);
  free(normals);
  free(texCoords);
  free(vertices);


}
